*======================================================================*
* WinBack extra routines, now all in one place.                        *
*         Copyright Norman Dunbar                                      *
*                   January 1992.                                      *
*======================================================================*
*                                                                      *
* x = DISC_space('flp1_') return free space in bytes on flp1_          *
*                                                                      *
*----------------------------------------------------------------------*
*====> NOTE multiply answer by 4 if checking a Miracle Hard Disc <=====*
*----------------------------------------------------------------------*
*                                                                      *
* x = FILE_length('flp1_file') returns length of flp1_file             *
* x = FILE_access('flp1_file') returns access byte of flp1_file        *
* x = FILE_type('flp1_file') return type byte of flp1_file             *
* x = FILE_data('flp1_file') returns dataspace of flp1_file            *
* x = FILE_unused('flp1_file') returns unused data of flp1_file        *
* x = FILE_update('flp1_file') returns update date of flp1_file        *
* x = FILE_reference('flp1_file') returns reference date of flp1_file  *
* x = FILE_backup('flp1_file') returns archive date of flp1_file       *
*                                                                      *
* err = FSET_access('flp1_file', data.byte) sets file access byte      *
* err = FSET_type('flp1_file', data.byte) sets file type byte          *
* err = FSET_data('flp1_file', data.long) sets files dataspace         *
* err = FSET_unused('flp1_file', data.long) sets unused part of header *
* err = FSET_update('flp1_file', data.long) sets file update date      *
* err = FSET_reference('flp1_file', data.long) sets file reference date*
* err = FSET_backup('flp1_file', data.long) sets file backup date      *
*                                                                      *
*----------------------------------------------------------------------*
* NOTE : Miracle hard discs use the top word of the reference date to  *
*        hold the file version number. The bottom word seems to be set *
*        to HEX 094A.                                                  *
*----------------------------------------------------------------------*
*                                                                      *
* BLACK,   BLACK%    returns 0                                         *
* BLUE,    BLUE%     returns 1      All of these are simply used       *
* RED,     RED%      returns 2      to avoid having to remember        *
* MAGENTA, MAGENTA%  returns 3      what number each colour is.        *
* GREEN,   GREEN%    returns 4                                         *
* CYAN,    CYAN%     returns 5                                         *
* WHITE,   WHITE%    returns 6                                         *
*----------------------------------------------------------------------*
*                                                                      *
* QDOS_ID = FOPEN_DIRECTORY(device$)                                   *
* ENTRY$  = FREAD_DIRECTORY$(QDOS_ID)                                  *
* CLOSE_DIRECTORY(QDOS_ID)                                             *
* if END_OF_DIRECTORY(QDOS_ID) then                                    *
*                                                                      *
* FOPEN_DIRECTORY returns a FLOAT which is the ACTUAL Qdos channel id  *
* or a negetive error code in the case of a failure to open.           *
*                                                                      *
* FREAD_DIRECTORY$ returns a 64 byte string corresponding to the next  *
* entry in the opened directory, or a two byte string holding the an   *
* error code in the second byte. This error code can be translated to  *
* the corresponding Qdos error code by subtracting 256 from its code & *
* then negating the result. END OF FILE returns 64 nulls.              *
*                                                                      *
* FCLOSE_DIRECTORY is a procedure to close the directory. The normal   *
* close routines cannot handle the closure, and if you leave it open   *
* all future accesses to the device will return 'in use'.              *
*                                                                      *
* END_OF_DIRECTORY returns 1 if EOF detected on the channel id or 0 if *
* not.                                                                 *
*----------------------------------------------------------------------*
*                                                                      *
* result% = FLOAD_FILE (#channel, num_bytes, address)                  *
* result% = FSAVE_FILE (#channel, num_bytes, address)                  *
*                                                                      *
* Attempt to load/save num_bytes characters from file #channel to/from *
* buffer at address. Returns a QDOS failure code or the number of      *
* bytes read/written.                                                  *
*                                                                      *
* result% = -01 --> Not complete                                       *
*           -06 --> Channel not open                                   *
*             0 --> End of file or zero bytes read into the buffer     *
*            >0 --> Number of bytes read into the buffer               *
*                                                                      *
* Can also return any other QDOS failure code.                         *
*----------------------------------------------------------------------*
*                                                                      *
* FORCE_ERROR error_code                                               *
*                                                                      *
* FORCE_ERROR does exactly that. It trips up Whatever is running at    *
* the time causing an error to abort the program, however, if there is *
* an error trap in use, then that will pick up the error and process   *
* it accordingly. This routine was written so that I could use one     *
* error trap in WinXxxx to handle errors, but, by using functions to   *
* perform various tasks, then I couldn't trap the errors as they       *
* occurred. What I had to do was test each result and duplicate some   *
* code. This procedure stops that duplication.                         *
*                                                                      *
* Error codes are in range -21 -> 0 -> 21 where positives are negated, *
* if an invalid error code is passed to the procedure it exits with an *
* err_sr, which I call ERR_Sinclair Research for obvious reasons, and  *
* returns a value of -25.                                              *
*                                                                      *
*----------------------------------------------------------------------*
*                                                                      *
* A function to convert a long integer into a string of ASCII digits   *
* separated by commas where appropriate. The maximum number it can     *
* handle is 2^31-2 as it is for UNSIGNED numbers only. Returns a 13    *
* character string with leading spaces.                                *
*                                                                      *
* PRINT COMMA$(2^31-2) = '2,147,483,646'                               *
* PRINT COMMA$(0)      = '            0'                               *
*                                                                      *
* Used by WinBack 2v00 onwards.                                        *
*----------------------------------------------------------------------*
*                                                                      *
* addr = POP_UP(#channel) returns address of saved contents buffer.    *
* POP_DOWN addr restores previous contents from buffer.                *
*----------------------------------------------------------------------*
*                                                                      *
* PRINT_ERROR$ #channel, error_code = print error text to #channel     *
* PRINT_ERROR$ error_code           = print error text to #1           *
*                                                                      *
*----------------------------------------------------------------------*
* TELL_MODE a function returnin 4 or 8 = current tv/monitor mode.      *
* NEWCHAN% a function which returns the next unused channel.           *
*======================================================================*


*======================================================================*
* AMENDMENTS.                                                          *
*                                                                      *
* 10th May 1991.                                                       *
* END_OF_DIRECTORY added to check for end of directory, as this removes*
* the need to return an error code for EOF, FREAD_DIRECTORY altered to *
* return a null string if any errors occur.                            *
*                                                                      *
* 2nd June 1991.                                                       *
* FILE_update & FILE_backup added.                                     *
*                                                                      *
* 22nd June 1991.                                                      *
* FILE_reference added.                                                *
*                                                                      *
* 4th July 1991.                                                       *
* Size of source reduced drastically from 17,986 by using a D5.W offset*
* to get the information from the buffer (Filedata_asm).               *
*                                                                      *
* FILE_update & FILE_backup were using the IOF_DATE trap #3 call and   *
* this only works for HARD DISC. New versions get info from the header.*
*                                                                      *
* 6th July 1991.                                                       *
* FSET_xxxx functions added. These return an error code or zero which  *
* is the result of attempting to set data in the file header. Also a   *
* correction was added to FILE_xxxx routines so that file is now closed*
* if an error is detected while reading file header. Mistakenly left   *
* open in earlier versions.                                            *
*                                                                      *
* 26th September 1991.                                                 *
* Fixed bug in FSET_TYPE & FSET_ACCESS where I was getting the value   *
* as a LONG onto the stack, and setting the data byte in the header to *
* the value of the HIGHEST byte and not the LOWEST so in 99% of cases  *
* I was actually setting the header byte to zero. Silly me.            *
*                                                                      *
* Changed FSET_any_date_routine to use IOF_DATE as FS_HEADS cannot set *
* the UPDATE, REFERENCE or BACKUP dates in a header, neither can DIYTK *
* routine SETHEAD (so there !). This now requires TK2 or a hard disc   *
* to work the dates.                                                   *
*                                                                      *
* 13th January 1992.                                                   *
* TELL_MODE added so that mode is only changed when required.          *
*======================================================================*


*======================================================================*
* Read in all the header files that are required & do other equates    *
*======================================================================*

           get     "win2_trap1_hdr"
           get     "win2_trap2_hdr"
           get     "win2_trap3_hdr"
           get     "win2_vectors_hdr"
           get     "win2_errors_hdr"
           get     "win2_sysvars_hdr"
           get     "win2_basic_hdr"
           get     "win2_windows_hdr"

length     equ     $00                 Offset to file_length in buffer
access     equ     $04                 Ditto to access byte
type       equ     $05                 Ditto to type byte
data       equ     $06                 Ditto to dataspace
unused     equ     $0A                 Ditto to unused part
update     equ     $34                 Ditto to update date
reference  equ     $38                 Ditto to reference date
backup     equ     $3C                 Ditto to backup date

me         equ    -$01                 This job
timeout    equ    -$01                 Timeout is forever
err_sr     equ    -$19                 Error Sinclair Research



*======================================================================*
* Call here to link all the extra routines in to SuperBasic.           *
*======================================================================*

start      move.w  bp_init,a2
           lea.l   define_1,a1
           jsr     (a2)                Link FILE_DATA

           lea.l   define_2,a1
           jsr     (a2)                Link DIRECTORY

           lea.l   define_3,a1
           jsr     (a2)                Link COLOURS

           lea.l   define_4,a1
           jsr     (a2)                Link FLOAD_FILE

           lea.l   define_5,a1
           jsr     (a2)                Link FORCE_ERROR

           lea.l   define_6,a1
           jsr     (a2)                Link COMMA

           lea.l   define_7,a1
           jsr     (a2)                Link POPUP

           lea.l   define_8,a1
           jsr     (a2)                Link PRINT_ERROR

           lea.l   define_9,a1
           jsr     (a2)                Link NEWCHAN

           lea.l   define10,a1
           jsr     (a2)                Link TELL_MODE

           rts                         Exit to Basic.

*----------------------------------------------------------------------*
* Data definition block for the FILE_DETAILS functions.                *
*----------------------------------------------------------------------*

define_1   dc.w    0                   No procedures
           dc.w    0                   End of procedures

           dc.w    (16+107+76+7)/8     Number of functions

           dc.w    d_space-*
           dc.b    10
           dc.b    'DISC_space'

           dc.w    d_space-*
           dc.b    10
           dc.b    'DISK_space'

           dc.w    f_length-*
           dc.b    11
           dc.b    'FILE_length'

           dc.w    f_access-*
           dc.b    11
           dc.b    'FILE_access'

           dc.w    f_type-*
           dc.b    9
           dc.b    'FILE_type'

           dc.w    f_data-*
           dc.b    9
           dc.b    'FILE_data'

           dc.w    f_unused-*
           dc.b    11
           dc.b    'FILE_unused'

           dc.w    f_ref-*
           dc.b    14
           dc.b    'FILE_reference'

           dc.w    f_updt-*
           dc.b    11
           dc.b    'FILE_update'

           dc.w    f_bkdt-*
           dc.b    11
           dc.b    'FILE_backup'

           dc.w    s_access-*
           dc.b    11
           dc.b    'FSET_access'

           dc.w    s_type-*
           dc.b    9
           dc.b    'FSET_type'

           dc.w    s_data-*
           dc.b    9
           dc.b    'FSET_data'

           dc.w    s_unused-*
           dc.b    11
           dc.b    'FSET_unused'

           dc.w    s_ref-*
           dc.b    14
           dc.b    'FSET_reference'

           dc.w    s_updt-*
           dc.b    11
           dc.b    'FSET_update'

           dc.w    s_bkdt-*
           dc.b    11
           dc.b    'FSET_backup'

           dc.w    0                   End of functions

*-----------------------------------------------------------------------*
* Data definition block for the DIRECTORY handling routines.            *
*-----------------------------------------------------------------------*

define_2   dc.w    (1+15+7)/8          1 procedure (long name)

           dc.w    close-*
           dc.b    15
           dc.b    'CLOSE_DIRECTORY'

           dc.w    0                   End of procedures

           dc.w    (2+15+16+16+7)/8    3 functions (long names)

           dc.w    fopen-*
           dc.b    15
           dc.b    'FOPEN_DIRECTORY'

           dc.w    fread-*
           dc.b    16
           dc.b    'FREAD_DIRECTORY$'

           dc.w    eof_dir-*
           dc.b    16
           dc.b    'END_OF_DIRECTORY'

           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* Data definition block for the COLOUR functions.                      *
*----------------------------------------------------------------------*

define_3   dc.w    0                   No procedures
           dc.w    0                   End of procedures

           dc.w    16                  16 functions

           dc.w    black_x-*
           dc.b    6
           dc.b    'BLACK%'

           dc.w    red_x-*
           dc.b    4
           dc.b    'RED%'

           dc.w    blue_x-*
           dc.b    5
           dc.b    'BLUE%'

           dc.w    magenta_x-*
           dc.b    8
           dc.b    'MAGENTA%'

           dc.w    green_x-*
           dc.b    6
           dc.b    'GREEN%'

           dc.w    cyan_x-*
           dc.b    5
           dc.b    'CYAN%'

           dc.w    white_x-*
           dc.b    6
           dc.b    'WHITE%'

           dc.w    yellow_x-*
           dc.b    7
           dc.b    'YELLOW%'

           dc.w    black_x-*
           dc.b    5
           dc.b    'BLACK'

           dc.w    red_x-*
           dc.b    3
           dc.b    'RED'

           dc.w    blue_x-*
           dc.b    4
           dc.b    'BLUE'

           dc.w    magenta_x-*
           dc.b    7
           dc.b    'MAGENTA'

           dc.w    green_x-*
           dc.b    5
           dc.b    'GREEN'

           dc.w    cyan_x-*
           dc.b    4
           dc.b    'CYAN'

           dc.w    white_x-*
           dc.b    5
           dc.b    'WHITE'

           dc.w    yellow_x-*
           dc.b    6
           dc.b    'YELLOW'

           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* DATA definition block for the FLOAD_FILE & FSAVE_FILE functions.     *
*----------------------------------------------------------------------*

define_4   dc.w    0                   No procedures
           dc.w    0                   End of procedures

           dc.w    (2+10+10+7)/8       2 functions (long names)
           dc.w    fload-*
           dc.b    10
           dc.b    'FLOAD_FILE'

           dc.w    fsave-*
           dc.b    10
           dc.b    'FSAVE_FILE'

           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* DATA definition block for the FORCE_ERROR procedure.                 *
*----------------------------------------------------------------------*

define_5   dc.w    (1+11+7)/8          1 procedure (long name)
           dc.w    f_error-*
           dc.b    11
           dc.b    'FORCE_ERROR'
           dc.w    0                   End of procedures

           dc.w    0                   No functions
           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* Data definition block for the COMMA$ function.                       *
*----------------------------------------------------------------------*

define_6   dc.w    0                   No procedures
           dc.w    0                   End of procedures

           dc.w    1                   One function

           dc.w    comma-*
           dc.b    6
           dc.b    'COMMA$'

           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* Data definition block for the POP_UP & POP_DOWN routines.            *
*----------------------------------------------------------------------*

define_7   dc.w    1                   One procedure

           dc.w    pop_down-*
           dc.b    8
           dc.b    'POP_DOWN'

           dc.w    0                   End of procedures
           dc.w    1                   One function

           dc.w    pop_up-*
           dc.b    6
           dc.b    'POP_UP'

           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* Data definition block for the PRINT_ERROR$ procedure.                *
*----------------------------------------------------------------------*

define_8   dc.w    (1+12+7)/8          One procedure (long name)

           dc.w    printerr-*
           dc.b    12
           dc.b    'PRINT_ERROR$'

           dc.w    0                   End of procedures
           dc.w    0                   No functions

           dc.w    0                   End of functions

*----------------------------------------------------------------------*
* DATA definition block the NEWCHAN% function.                         *
*----------------------------------------------------------------------*

define_9   dc.w  0                     No procedures
           dc.w  0                     End of procedures

           dc.w  1                     1 function
           dc.w  newchan-*
           dc.b  8
           dc.b  'NEWCHAN%'

           dc.w  0                     End of functions

*----------------------------------------------------------------------*
* DATA definition block the NEWCHAN% function.                         *
*----------------------------------------------------------------------*

define10   dc.w    0                   No procedures
           dc.w    0                   End of procedures

           dc.w    (1+9+7)/8           1 Function (with a long name)
           dc.w    tellmode-*
           dc.b    9
           dc.b    'TELL_MODE'

           dc.w    0                   End of functions


*----------------------------------------------------------------------*
* DISC_space & DISK_space.                                             *
*----------------------------------------------------------------------*

d_space    move.l  bv_rip(a6),d7       Store maths stack in a TURBO like manner
           bsr     get_1_str           Fetch only 1 string
           bne.s   fdb_param           Must be zero or an error occurred
           moveq   #4,d3               It the directory we want
           bsr     open_file           Open the directory
           bne     ret_err             Bale out
           moveq   #fs_mdinf,d0        Prepare to get media information
           moveq   #timeout,d3         Timeout is all day
           lea.l   buffer,a1           For media name (not used)
           trap    #3                  Get info
           tst.l   d0                  Did it work ?
           bne     ret_err             No
           move.w  #0,d1               Clear low word
           swap    d1                  D1.L = sectors free
           asl.l   #8,d1               Multiply by 256
           asl.l   #1,d1               Then by 2 giving 512 = bytes free
           move.l  d1,-(a7)            Store result
           bsr     close_file          Close directory
           move.l  (a7)+,d1            Retrieve result
           bra     set_math            Return the value.

*----------------------------------------------------------------------*
* FILE_length.                                                         *
*----------------------------------------------------------------------*

f_length   moveq   #length,d5
           bra.s   get_it

*----------------------------------------------------------------------*
* FILE_access.                                                         *
*----------------------------------------------------------------------*

f_access   moveq   #access,d5
           bra.s   get_it

*----------------------------------------------------------------------*
* FILE_type.                                                           *
*----------------------------------------------------------------------*

f_type     moveq   #type,d5
           bra.s   get_it

*----------------------------------------------------------------------*
* FILE_data.                                                           *
*----------------------------------------------------------------------*

f_data     moveq   #data,d5
           bra.s   get_it

*----------------------------------------------------------------------*
* FILE_unused.                                                         *
*----------------------------------------------------------------------*

f_unused   moveq   #unused,d5
           bra.s   get_it

fdb_param  moveq   #err_bp,d0
           bra     ret_err

bad_device moveq   #err_bn,d0
           bra     ret_err

*----------------------------------------------------------------------*
* FILE_reference.                                                      *
*----------------------------------------------------------------------*

f_ref      moveq   #reference,d5
           bra.s   get_it

*----------------------------------------------------------------------*
* FILE_update.                                                         *
*----------------------------------------------------------------------*

f_updt     moveq   #update,d5
           bra.s   get_it

*----------------------------------------------------------------------*
* FILE_backup.                                                         *
*----------------------------------------------------------------------*

f_bkdt     moveq   #backup,d5

get_it     move.l  bv_rip(a6),d7       Store stack
           bsr.s   get_open            Open, read heads & close
           bne.s   ret_err             Something didn't work
           lea.l   buffer,a1           Start of header
           cmpi.w  #access,d5          Byte size required ?
           beq.s   get_byte            Yes
           cmpi.w  #type,d5            Try the other byte size
           bne.s   get_long            No, must be long

get_byte   move.b  0(a1,d5.w),d1       Get a byte of data
           bra     set_math            Return it

get_long   move.l  0(a1,d5.w),d1       Get a long word of data
           bra     set_math            Return it

*----------------------------------------------------------------------*
* Subroutine to get 1 string onto the stack, sets Z if ok otherwise Z  *
* is unset.                                                            *
*----------------------------------------------------------------------*

get_1_str  move.w  ca_gtstr,a2         Prepare to get strings
           jsr     (a2)                Get them
           bne.s   get_1_exit          Oops
           subq.w  #1,d3               Check we got only 1
           beq.s   get_1_exit          Yes
           moveq   #err_bp,d0          No

get_1_exit rts

*----------------------------------------------------------------------*
* Subroutine to get 1 string onto the stack, open the file, read the   *
* file header into the buffer. Sets Z if all ok, unset otherwise.      *
*----------------------------------------------------------------------*

get_open   bsr.s   get_1_str           Get filename
           bne.s   get_1_exit          Bale out if errors
           moveq   #1,d3               Old shared file
           bsr.s   open_file           Open it
           bne.s   get_1_exit          Bale out if errors
           bsr.s   read_head           Read file header
           move.l  d0,-(a7)            Store error if any
           bsr.s   close_file          Close file
           move.l  (a7)+,d0            Retrieve error code
           rts

*----------------------------------------------------------------------*
* Subroutine to open a file, filename is on the stack.                 *
*----------------------------------------------------------------------*

open_file  tst.w   0(a6,a1.l)          Check size word
           bne.s   len_ok

           moveq   #err_bn,d0          Bad name
           rts

len_ok     moveq   #io_open,d0         Prepare to open file
           moveq   #me,d1              For this job
           move.l  a1,a0               Address of filename
           trap    #4                  Unrelative it
           trap    #2                  Open it
           tst.l   d0
           rts                         Return Z set if ok

*----------------------------------------------------------------------*
* Subroutine to read the header of the file on channel (A0).           *
*----------------------------------------------------------------------*

read_head  moveq   #fs_headr,d0        Prepare to read header
           moveq   #64,d2              Buffer length
           moveq   #timeout,d3         Take all day
           lea.l   buffer,a1           Where to put the header
           trap    #3                  Read file header
           tst.l   d0                  Check for errors
           rts                         Return Z set if ok

*----------------------------------------------------------------------*
* Subroutine to close the file on channel (A0).                        *
*----------------------------------------------------------------------*

close_file moveq   #io_close,d0        Prepare to close it
           trap    #2                  Close it
           moveq   #0,d0               Close does not fail (!)
           rts                         Exit

*----------------------------------------------------------------------*
* Return an error code currently stored in D0.L by moving it to D1.L   *
* and converting it to floating point.                                 *
*----------------------------------------------------------------------*

ret_err    move.l  d0,d1               Put error code into D1 for conversion

*----------------------------------------------------------------------*
* Set up the maths stack to its entry value, get 6 bytes on it for a   *
* floating point result, convert D1.L to float and exit to basic with  *
* the result.                                                          *
*----------------------------------------------------------------------*

set_math   move.l  d7,a1               Its Turbo time again
           move.l  a1,bv_rip(a6)       Store it
           move.l  d1,-(a7)            Stack result for now
           moveq   #6,d1               We need 6 bytes for a float
           move.w  bv_chrix,a2         Get some room
           jsr     (a2)                (No errors)
           move.l  (a7)+,d1            And the result code again
           move.l  bv_rip(a6),a1       Get the (new) stack pointer
           bsr     D1TOFP              Convert D1.L to floating point

ret_fp     subq.l  #$06,a1             Space for 1 float
           move.l  d5,$02(a1,a6.l)     Store the mantissa
           move.w  d4,$00(a1,a6.l)     Store the exponent
           moveq   #$02,d4             Signal float result
           move.l  a1,bv_rip(a6)       Store new stack
           moveq   #$00,d0             No errors
           rts                         Exit to basic with the result

*----------------------------------------------------------------------*
* Buffer for file headers etc.                                         *
*----------------------------------------------------------------------*

buffer     ds.w    32                  File header buffer

*----------------------------------------------------------------------*
* Now all the FSET routines.                                           *
*----------------------------------------------------------------------*

*----------------------------------------------------------------------*
* FSET_type                                                            *
*----------------------------------------------------------------------*

s_type     moveq   #type,d5
           bra.s   set_all

*----------------------------------------------------------------------*
* FSET_access                                                          *
*----------------------------------------------------------------------*

s_access   moveq   #access,d5
           bra.s   set_all

*----------------------------------------------------------------------*
* FSET_data                                                            *
*----------------------------------------------------------------------*

s_data     moveq   #data,d5
           bra.s   set_all

*----------------------------------------------------------------------*
* FSET_unused                                                          *
*----------------------------------------------------------------------*

s_unused   moveq   #unused,d5
           bra.s   set_all

*----------------------------------------------------------------------*
* FSET_update                                                          *
*----------------------------------------------------------------------*

s_updt     moveq   #update,d5
           bra.s   set_all

*----------------------------------------------------------------------*
* FSET_reference                                                       *
*----------------------------------------------------------------------*

s_ref      moveq   #reference,d5
           bra.s   set_all

*----------------------------------------------------------------------*
* FSET_backup                                                          *
*----------------------------------------------------------------------*

s_bkdt     moveq   #backup,d5

set_all    move.l  bv_rip(a6),d7       Store maths pointer Turbo style
           lea.l   $10(a3),a0          Check for two parameters
           cmpa.l  a0,a5
           bne     bad_param           Oops...
           subq.l  #8,a5               Only one string to get
           bsr     get_1_str
           bne     ret_err             Return error code
           moveq   #0,d3               Old exclusive file
           bsr     open_file           Open it
           bne     ret_err             Oops...
           bsr     read_head           Read current header
           bne.s   err_close           Oops...
           move.l  d7,a1               Reset maths stack pointer
           move.l  d7,bv_rip(a6)       Just in case
           move.l  a5,a3
           addq.l  #8,a5               Only one long int to get
           move.l  a0,-(a7)            Store channel id
           move.w  ca_gtlin,a2         Prepare for long integers
           jsr     (a2)                Corrupts a0
           move.l  (a7)+,a0            Retrieve channel id
           tst.l   d0                  Check flags
           bne     ret_err             Didn't get the data
           subq.w  #1,d3               Check only one
           bne     bad_param           Oops...
           lea.l   buffer,a2           Point to header
           cmpi.w  #length,d5          Long sized ?
           beq.s   set_long            Yes
           cmpi.w  #type,d5            Byte sized ?
           bhi.s   not_byte            No

set_byte   move.b  3(a6,a1.l),0(a2,d5.w) Only get the lowest byte
           bra.s   set_head            Store byte & set header

not_byte   cmpi.w  #update,d5          Need to use extended drivers ?
           bcc.s   extended            Yes

set_long   move.l  0(a6,a1.l),0(a2,d5.w)

set_head   moveq   #fs_heads,d0        Prepare to set header (A0.L holds ID)
           moveq   #timeout,d3         Take all day
           exg     a1,a2               Point A1 to buffer & save maths stack
           trap    #3
           exg     a1,a2               Get stack back

err_close  move.l  d0,-(a7)            Store error code if any
           bsr     close_file          Close the file (no errors)
           move.l  (a7)+,d0            Get error code
           bra     ret_err             Return 0 or error code

extended   subi.w  #update,d5          D5 is 0, 4 or 8 now
           asr.w   #2,d5               D5 is 0, 1 or 2 for UPDT, REF or BKDT
           moveq   #iof_date,d0        Prepare to set header (A0.L holds ID)
           move.l  0(a6,a1.l),d1       Get the required date
           move.b  d5,d2               Get 'which date' parameter
           moveq   #timeout,d3         Can take a while this, you know...
           trap    #3                  Do extended header set
           bra.s   err_close           Close channel & return error (if any)



*----------------------------------------------------------------------*
* Code for the colour functions.                                       *
*----------------------------------------------------------------------*

black_x    moveq   #0,d7
           bra.s   ret_d7

blue_x     moveq   #1,d7
           bra.s   ret_d7

red_x      moveq   #2,d7
           bra.s   ret_d7

magenta_x  moveq   #3,d7
           bra.s   ret_d7

green_x    moveq   #4,d7
           bra.s   ret_d7

cyan_x     moveq   #5,d7
           bra.s   ret_d7

yellow_x   moveq   #6,d7
           bra.s   ret_d7

white_x    moveq   #7,d7

ret_d7     moveq   #2,d1               Space for a word integer required
           move.w  bv_chrix,a2         Prepare to get some space
           jsr     (a2)                Go get it (no errors)
           move.l  bv_rip(a6),a1       Get new maths pointer
           subq.l  #2,a1               Make room
           move.w  d7,0(a6,a1.l)       Store colour or mode value
           move.l  a1,bv_rip(a6)       Store updated maths pointer
           moveq   #3,d4               Result is word integer
           moveq   #0,d0               No errors
           rts                         Back to the Basic(s)

*----------------------------------------------------------------------*
* TELL_MODE                                                            *
*----------------------------------------------------------------------*

tellmode   moveq   #MT_DMODE,d0
           moveq   #-1,d1              Read mode
           moveq   #-1,d2              Read display (don't care)
           trap    #1                  Do it
           tst.b   d1                  Check mode
           bne.s   mode_8              Mode is 8

mode_4     moveq   #4,d7               Return 4 for mode 4
           bra.s   ret_d7

mode_8     moveq   #8,d7               Return 8 for mode 8
           bra.s   ret_d7

*----------------------------------------------------------------------*
* CLOSE_DIRECTORY(QDOS_ID)                                             *
*----------------------------------------------------------------------*

close      move.w  ca_gtlin,a2         Parameters are long words
           jsr     (a2)                Get them
           bne     exit                Oops....
           subq.w  #1,d3               Should be 1 only
           bne     bad_param           Ooops....
           moveq   #io_close,d0        We want to close a channel
           move.l  0(a6,a1.l),a0       Id is now in A0
           trap    #2                  Close it (no errors)
           rts                         But just in case

*----------------------------------------------------------------------*
* QDOS_ID = FOPEN_DIRECTORY(device$)                                   *
*----------------------------------------------------------------------*

fopen      move.l  bv_rip(a6),d7       If Turbo can do it, so can I
           move.w  ca_gtstr,a2         We want some strings
           jsr     (a2)                So get them
           bne     exit                Oops....
           subq.w  #1,d3               Should only be 1
           bne     bad_param           Oops....
           moveq   #io_open,d0         We want to open a channel
           moveq   #me,d1              For this job
           moveq   #4,d3               And it will be a directory
           move.l  a1,a0               Pointer to name
           trap    #4                  Unrelative it
           trap    #2                  Open it
           tst.l   d0                  Check return code
           beq.s   open_ok             It opened ok
           bra     ret_err             Return error code

open_ok    move.l  a0,d1               Return the Qdos id
           bra     set_math            Convert it & exit

*----------------------------------------------------------------------*
* ENTRY$ = FREAD_DIRECTOR$(QDOS_ID)                                    *
*----------------------------------------------------------------------*

fread      move.w  ca_gtlin,a2         We need a long word
           jsr     (a2)                Go get it
           bne     exit                Oops....
           subq.w  #1,d3               But only 1
           bne     bad_param           Oops....
           move.l  0(a6,a1.l),a0       It is a channel id
           addq.l  #4,a1               Clear stack
           move.l  a1,bv_rip(a6)       Store top of stack
           moveq   #66,d1              We should need 2 + 64 bytes on the stack
           move.w  bv_chrix,a2
           jsr     (a2)                Reserve the space
           move.l  bv_rip(a6),a1       Get (new) stack pointer
           suba.l  #66,a1              Make the room
           move.l  a1,-(a7)            Store the pointer
           addq.l  #2,a1               Point to buffer space (on stack)
           moveq   #io_fstrg,d0        We want to read some bytes
           moveq   #64,d2              And 64 seems like a good number
           moveq   #timeout,d3         It can take all day to do it
           trap    #4                  Unrelative the buffer pointer
           trap    #3                  Read the bytes
           move.l  (a7)+,a1            Get the stack pointer
           tst.l   d0                  Check for errors
           beq.s   read_ok             There were none
           adda.l  #64,a1              Tidy unwanted stack space
           clr.w  0(a6,a1.l)           Return null length string
           bra.s   ret_str             And return it

read_ok    move.w  #64,0(a6,a1.l)      Store the string size

ret_str    move.l  a1,bv_rip(a6)       Store the stack pointer
           moveq   #0,d0               No errors
           moveq   #1,d4               Result is a string
           rts                         And exit

*----------------------------------------------------------------------*
* END_OF_DIRECTORY(QDOS_ID)                                            *
*----------------------------------------------------------------------*

eof_dir    move.w  ca_gtlin,a2         Parameters are long words
           jsr     (a2)                Get them
           bne.s   exit_eof            Oops....
           subq.w  #1,d3               Should be 1 only
           bne     bad_param           Ooops....
           moveq   #io_pend,d0         We want to close a channel
           moveq   #timeout,d3         Timeout
           move.l  0(a6,a1.l),a0       Id is now in A0
           move.l  a1,d7               Store stack, it get corrupted next
           trap    #3                  Check for EOF
           move.l  d7,a1               Get stack back
           tst.l   d0                  Check for errors
           beq.s   ret_not             Not end of file
           cmpi.l  #err_ef,d0          Was error EOF ?
           bne.s   exit_eof            Oh dear

ret_eof    move.w  #1,2(a6,a1.l)       Store a 1
           bra.s   return

ret_not    clr.w   2(a6,a1.l)          Store a zero

return     addq.l  #2,a1               Set up stack
           move.l  a1,bv_rip(a6)       Store stack
           moveq   #3,d4               Return integer
           moveq   #0,d0               Clear EOF error code

exit_eof   rts                         Back to basic



*----------------------------------------------------------------------*
* CODE.                                                                *
*----------------------------------------------------------------------*

fload      movea.l #io_fstrg,a4
           bra.s   do_it_all

fsave      movea.l #io_sstrg,a4

*----------------------------------------------------------------------*
* First check that only three parameters are expected and the first of *
* them has a hash in front.                                            *
*----------------------------------------------------------------------*

do_it_all  lea.l   $18(a3),a0                    Check for 3 parameters
           cmpa.l  a0,a5
           bne     bad_param                     Oops....

found_3    btst    #7,1(a6,a3.l)                 First should have a #
           beq     bad_param                     Oops....

*----------------------------------------------------------------------*
* Now get the first two, channel number & number of bytes, as word     *
* integers and make sure that only 2 parameters were actually obtained *
*----------------------------------------------------------------------*

hash_ok    subq.l  #8,a5                         Force 2 params only
           move.w  ca_gtint,a2                   Get 2 word integers
           jsr     (a2)                          Do it
           bne     exit                          Oops....
           subq.w  #2,d3                         Make sure we got 2
           bne     bad_param                     Oops....

*----------------------------------------------------------------------*
* Put num_bytes into D7.W and channel id into D5.W as these are not    *
* corrupted by the parameter fetching routines. D5.W as a channel      *
* number is converted into a QDOS channel id in D5.L                   *
*----------------------------------------------------------------------*

check_id   moveq   #0,d5                         Need to clear top word of D5
           move.w  0(a6,a1.l),d5                 Get channel number
           blt     bad_param                     Cannot be negative
           mulu    #40,d5                        Convert into an offset
           add.l   bv_chbas(a6),d5               Add base
           cmp.l   bv_chp(a6),d5                 Still in table ?
           bge.s   not_open                      No, oh dear
           move.l  0(a6,d5.l),d5                 Yes, get channel id & set flags
           bge.s   chan_open                     Negatives are closed

not_open   moveq   #err_no,d0                    Channel not open
           rts                                   Back to SuperBasic


*----------------------------------------------------------------------*
* Channel D5 is open, so fetch the number of bytes into D7.W, negative *
* amounts are not allowed. The maths stack is then adjusted and the    *
* buffer address is fetched as a long word integer. A check is made to *
* ensure that only 1 parameter is fetched.                             *
*----------------------------------------------------------------------*

chan_open  move.w  2(a6,a1.l),d7                 D7.w = num_bytes
           blt     bad_param                     Negative is bad
           addq.l  #4,a1                         Adjust stack over 2 word ints
           move.l  a1,bv_rip(a6)                 Store maths stack
           move.l  a5,a3                         Only 1 param to get
           addq.l  #8,a5
           move.w  ca_gtlin,a2                   And it's a long word
           jsr     (a2)                          Go get it
           bne     exit                          Oops....
           subq    #1,d3                         Check only 1
           bne     bad_param                     Oops....

*----------------------------------------------------------------------*
* Now, get the channel id in A0.L, store the maths stack again, and    *
* get the number of bytes to load into D2.W which, if zero will set    *
* the zero flag and we will simply return, otherwise load the required *
* number of bytes into the buffer at A1.L                              *
*----------------------------------------------------------------------*

           move.l  d5,a0                         Channel id
           move.l  a1,bv_rip(a6)                 Back into store
           move.l  0(a6,a1.l),a1                 Buffer address
           move.w  d7,d2                         Num_bytes
           bne.s   do_load                       Don't bother if num_bytes = zero
           moveq   #0,d0                         Just return zero
           bra.s   load_zero

do_load    move.l  a4,d0                         IO_FSTRG or IO_SSTRG
           moveq   #-1,d3                        Timeout is all day if you want
           trap    #3                            Do it

*----------------------------------------------------------------------*
* Either all went well, in which case the buffer is full of data or an *
* error code is in D0.L. If an error has occurred, return the error    *
* code in D0.L otherwise return the number of bytes read in which is   *
* currently held in D1.W                                               *
*----------------------------------------------------------------------*

           cmpi.l  #err_ef,d0                    Was it EOF (who cares for FSAVE)
           beq.s   ret_d1                        Yes, return bytes read in
           tst.l   d0                            Check error code
           bne.s   load_zero                     Error, return its code

ret_d1     move.w  d1,d0                         Return number of bytes read

load_zero  move.l  bv_rip(a6),a1                 Get maths stack again
           addq.l  #2,a1                         Adjust to 1 word int only
           move.w  d0,0(a6,a1.l)                 Put the result on the stack
           moveq   #0,d0                         No errors
           moveq   #3,d4                         Word integer result
           move.l  a1,bv_rip(a6)                 Store new maths stack
           rts                                   Go home




*----------------------------------------------------------------------*
* FORCE_ERROR qdos_error_code                                          *
*----------------------------------------------------------------------*

f_error    move.w  ca_gtlin,a2         Parameters are long words
           lea.l   8(a3),a5            Make sure only one gets through
           jsr     (a2)                Get them
           bne     exit                Oops....
           move.l  0(a6,a1.l),d0       Error code in D0.L
           beq     exit                Bale out early if zero
           blt.s   range               Negatives are ok
           neg.l   d0                  Positives are negatives now

range      cmpi.l  #err_bl,d0          In range -21 to -1 ?
           bge     exit                Yes, exit

silly      moveq   #err_sr,d0          Error Sinclair research
           rts                         Back to Basic



*----------------------------------------------------------------------*
* CODE. (For the function).                                            *
*----------------------------------------------------------------------*

comma      move.w  ca_gtlin,a2         Long integer required
           jsr     (a2)                Go get it
           bne     exit                Bale out if errors
           subq.w  #1,d3               Check how many
           bne     BAD_PARAM           Only one is allowed
           move.l  0(a6,a1.l),d4       Get the value to convert
           moveq   #12,d1              4 + 12 = 16 (we need 15)
           move.l  a1,bv_rip(a6)       Just in case
           move.w  bv_chrix,a2         request for space
           jsr     (a2)                Get the space
           move.l  bv_rip(a6),a1       Get new pointer
           suba.l  #10,a1              Point to buffer space
           move.l  a1,a0               Buffer pointer is A0.L (relative A6)
           moveq   #12,d0              13 bytes

space_fill move.b  #' ',0(a6,a0.l)     Space fill one byte in buffer
           addq.l  #1,a0               Point to next byte in buffer
           dbra    d0,space_fill       And do the rest
           move.b  #0,0(a6,a0.l)       Store a null padding byte
           move.l  a1,a0               Get the buffer start address again
           subq.l  #2,a1               Point to base of maths stack

convert    lea.l   tens_table,a2       Point to the tens table
           moveq   #0,d1               Total number of digits in number

next_digit move.l  (a2)+,d2            Read a power of ten
           beq.s   all_done            Zero means we are done
           clr.b   d0                  This digit

digit_loop sub.l   d2,d4               Subtract current power of ten
           blt.s   buff_digit          Gone below zero, too far
           addq.b  #1,d0               Still ok, increment this digit
           bra.s   digit_loop          And try again

buff_digit add.l   d2,d4               Add current power of ten back on
           tst.b   d0                  This digit non zero ?
           bne.s   not_a_zero          Yes
           tst.w   d1                  No, any digits yet ?
           bne.s   not_a_zero          Yes, do this zero digit

do_space   addq.l  #1,a0               Miss out leading zeros
           bra.s   test_comma          But check if a comma goes here

not_a_zero addi.b  #'0',d0             Convert to ASCII
           move.b  d0,0(a6,a0.l)       Store in the buffer
           addq.l  #1,a0               Point to next free byte
           addq.w  #1,d1               Increment total digits counter

test_comma lea.l   ten_8,a3            First place a comma will go
           cmpa.l  a3,a2               Test to see if one is needed
           beq.s   do_comma            Yes it is
           adda.l  #12,a3              Points to 'ten_5'
           cmpa.l  a3,a2               Test again
           beq.s   do_comma            Yes, comma required
           adda.l  #12,a3              Points to 'ten_2'
           cmpa.l  a3,a2               Last try
           bne.s   next_digit          No comma

do_comma   tst.w   d1                  Check for any digits yet
           beq.s   no_digits           None yet
           move.b  #',',0(a6,a0.l)     Put a comma in the buffer

no_digits  addq.l  #1,a0               Point to next free byte
           bra.s   next_digit          Go do the next digit

all_done   tst.w   d1                  Any digits at all
           bne.s   some_digit          Yes
           move.b  #'0',-1(a6,a0.l)    Store a zero

some_digit move.w  #13,0(a6,a1.l)      Finished, store word count
           bra     ret_str

tens_table dc.l    1000000000          10 ^ 9
ten_8      dc.l     100000000          10 ^ 8
           dc.l      10000000          10 ^ 7
           dc.l       1000000          10 ^ 6
ten_5      dc.l        100000          10 ^ 5
           dc.l         10000          10 ^ 4
           dc.l          1000          10 ^ 3
ten_2      dc.l           100          10 ^ 2
           dc.l            10          10 ^ 1
           dc.l             1          10 ^ 0
           dc.l             0          End of table



*----------------------------------------------------------------------*
* POP_DOWN addr.                                                       *
*----------------------------------------------------------------------*

pop_down   move.w  ca_gtlin,a2         Long integer required
           jsr     (a2)                Go get it
           bne     exit                Oops...
           subq.w  #1,d3               There can be only one (Highlander !)
           bne     BAD_PARAM           Oops...

param_ok   move.l  0(a6,a1.l),a0       Start address of buffer
           move.l  a0,a2               Use A2 as buffer pointer
           move.l  (a2)+,d0            Get marker word
           cmpi.l  #'BUGr',d0          Is this a marker
           bne     bad_param           No, bale out
           move.l  (a2)+,a3            Screen address (destination)
           move.w  (a2)+,d4            Number of bytes per screen line
           move.w  (a2)+,d5            Number of lines to be restored

*----------------------------------------------------------------------*
* A2.L = From address                                                  *
* A3.L = To address                                                    *
* D4.W = Bytes in one line to be restored                              *
* D5.W = Number of lines to restore                                    *
*----------------------------------------------------------------------*

           subq.w  #1,d4               For DBRA
           subq.w  #1,d5               For DBRA

pd_outer   move.w  d4,d3               Use D3 as byte count
           move.l  a3,a5               Store address of this screen line

pd_inner   move.b  (a2)+,(a3)+         Restore one byte
           dbra    d3,pd_inner         Restore this screen line
           move.l  a5,a3               Get address of this screen line
           adda.l  #128,a3             Point to next screen line
           dbra    d5,pd_outer         And restore all the lines we require

reclaim    moveq   #mt_rechp,d0        Finished with buffer
           trap    #1                  Reclaim it
           moveq   #0,d0               No errors
           rts                         Back to Basic

*----------------------------------------------------------------------*
* addr = POP_UP(#channel).                                             *
*----------------------------------------------------------------------*

pop_up     move.l  bv_rip(a6),d7       Store maths stack TURBO style
           move.w  ca_gtint,a2         Word integer required
           jsr     (a2)                Go get it
           bne     exit                Oops...
           subq.w  #1,d3               There can be only one (Highlander 2 !)
           bne     BAD_PARAM           Oops...
           move.w  0(a6,a1.l),d1       Channel number
           bsr     D1TOCHANID          Convert d1 to Qdos id in A0
           bne     ret_err             Return error code
           moveq   #sd_extop,d0        Prepare to use the channel
           moveq   #timeout,d3         Take all the time in the world
           lea.l   extop,a2            The routine to call
           trap    #3                  Do it
           tst.l   d0                  Did it work ?
           bne     ret_err             No

*----------------------------------------------------------------------*
* A1.L = base of channel definition block                              *
*----------------------------------------------------------------------*

           move.l  a1,a0               Use A0 for offsets
           bra.s   first_col           Jump over extop routine

extop      move.l  a0,a1               Pass back channel definition address
           moveq   #0,d0               No errors
           rts                         Finished

*----------------------------------------------------------------------*
* Calculate offset to first column of window to be saved. This is the  *
* number of pixels across to the x_origin DIV 8 to give bytes then     *
* times 2 and made even. (One word of memory defines 8 pixels on the   *
* monitor/tv screen.                                                   *
*----------------------------------------------------------------------*

first_col  moveq   #0,d5               Clear First column offset in D5
           move.w  sd_xmin(a0),d5      Get x_origin of window
           move.l  d5,d4               Needed later
           asr.w   #3,d5               Get DIV 8 of x_origin
           asl.w   #1,d5               Times 2 and even (faster than BCLR)

*----------------------------------------------------------------------*
* D5.W = offset from start of screen memory to first pixel of window.  *
*                                                                      *
* Now, calculate the length in bytes of each screen line in the window *
* which is (((pixel width + x_origin + 6) DIV 8) * 2) - x_origin in    *
* bytes as calculated above.                                           *
*----------------------------------------------------------------------*

how_wide   add.w   sd_xsize(a0),d4     Width of window + start of window
           addq.w  #6,d4
           asr.w   #3,d4               get DIV 8 of width
           asl.w   #1,d4               Times 2 and even (faster than BCLR)
           sub.w   d5,d4               Subtract first column offset

*----------------------------------------------------------------------*
* D4.W = Number of bytes per screen line to be saved                   *
*                                                                      *
* Now, calculate the required amount of memory to save the window,     *
* which is the height in pixels * width in bytes (as above) + 8 for    *
* storage overheads (one long + two words). when calculated, allocate  *
* a heap to hold it in.                                                *
*----------------------------------------------------------------------*

how_much   move.l  d4,d1               D1.L = bytes per screen line
           mulu    sd_ysize(a0),d1     Times number of lines in window
           add.l   #12,d1              Add overhead
           move.l  a0,-(a7)            Store channel definition address
           moveq   #mt_alchp,d0        Prepare for getting some heap
           moveq   #me,d2              For this job
           trap    #1                  Get some heap
           move.l  a0,a1               Store heap address
           move.l  (a7)+,a0            Get definition address
           tst.l   d0                  Check for errors
           bne     ret_err             Return error code

*----------------------------------------------------------------------*
* Now calculate the start address of the screen memory which is the    *
* address that will be the top right corner of the window to be saved. *
* This is (128 * y_origin) + screen start address + offset to first    *
* column as above.                                                     *
*                                                                      *
* Note that screen address is obtained from the channel definition as  *
* the THOR/MINERVA may not have the screen at $20000 (131072) !!!!!!!  *
*----------------------------------------------------------------------*

scr_addr   moveq   #0,d0               Just in case
           move.w  sd_ymin(a0),d0      Get y_origin of window
           asl.l   #7,d0               Times 128
           add.l   d5,d0               Add in first column offset
           add.l   sd_scrb(a0),d0      Add base of screen (THOR/MINERVA)

*----------------------------------------------------------------------*
* Put the expected result into D1.L for return to Basic (the buffer    *
* address) then store screen_address.L bytes_wide.W & pixels_deep.W in *
* the buffer at the start. The rest of the buffer will hold the data   *
* from the window.                                                     *
*----------------------------------------------------------------------*

           move.l  a1,d1               Function result in D1.L
           move.l  #'BUGr',(a1)+       Store marker for saved window
           move.l  d0,(a1)+            Store screen_address
           move.w  d4,(a1)+            Store bytes per line
           move.w  sd_ysize(a0),d5     Get number of lines in window
           move.w  d5,(a1)+            And store it
           move.l  d0,a3               Source address is screen_address

*----------------------------------------------------------------------*
* A0.L = Address of channel definition block                           *
* A1.L = Destination address (save buffer)                             *
* A3.L = Source address (screen)                                       *
* D1.L = Save buffer address = -8(A1)                                  *
* D4.W = Number of bytes in one line                                   *
* D5.W = Number of lines in window                                     *
*----------------------------------------------------------------------*

           subq.w  #1,d4               Adjust for DBRA
           subq.w  #1,d5               Ditto

pu_outer   move.w  d4,d3               Use D3 as DBRA counter
           move.l  a3,a5               Store source address

pu_inner   move.b  (a3)+,(a1)+         Store one byte
           dbra    d3,pu_inner         Do whole of one line
           move.l  a5,a3               Get address of this screen line
           adda.l  #128,a3             Point to next screen line
           dbra    d5,pu_outer         Do the next screen line
           bra     set_math

*----------------------------------------------------------------------*
* PRINT_ERROR$ [ #channel , ] error_code                               *
*----------------------------------------------------------------------*

printerr   move.w  ca_gtlin,a2         Long integer(s) required
           jsr     (a2)                Go get it/them
           bne.s   pe_exit             Oops...
           move.l  a1,a3               Save maths stack pointer
           cmpi.w  #1,d3               Try only one...
           beq.s   got_1               Yes, use default channel
           cmpi.w  #2,d3               Try two...
           beq.s   get_2               Yes, channel supplied
           bra     BAD_PARAM           Oops...

got_1      moveq   #1,d1               Default channel is #1
           subq.l  #4,a3               Fiddle stack pointer
           bra.s   get_id              Calculate channel id

get_2      move.l  0(a6,a3.l),d1       Get channel id from stack
           bmi     BAD_PARAM           negative channels are not good

get_id     bsr     D1TOCHANID          Convert to qdos id in A0
           bne.s   pe_exit             Errors = return to SuperBasic

got_id     move.l  4(a6,a3.l),d0       Get error code
           bpl     BAD_PARAM           Positive is bad
           cmpi.l  #err_bl,d0          Check in range
           blt     BAD_PARAM           Too far negative to be an error code

open       move.w  ut_err,a2           Get the vector
           jsr     (a2)                Print error text
           moveq   #0,d0               There are no errors

pe_exit    rts                         Back to Basic

*-----------------------------------------------------------*
* Code for NEWCHAN% starts here                             *
*-----------------------------------------------------------*

newchan    moveq    #0,d4           start off at channel 0
           move.l   bv_chbas(a6),a2 start of channel table
           move.l   bv_chp(a6),a3   end of channel table

try_ch     cmpa.l   a3,a2          finished yet ?
           bcc.s    got_ch         yes - d4 is next channel to use
           tst.b    0(a2,a6.l)     is channel in use ?
           bmi.s    got_ch         no - d4 is next channel to use
           addq.w   #1,d4          yes - try next one
           lea.l    40(a2),a2      point to next entry in channel table
           bra.s    try_ch         and try it

got_ch     moveq    #2,d1          2 bytes required for an integer result
           move.w   bv_chrix,a2    reserve space on arithmetic stack
           jsr      (a2)           do it
           move.l   bv_rip(a6),a1  get (possible) new value for arithmetic stack
           subq.l   #2,a1          and claim our reserved space

ret_d4     move.w   d4,0(a1,a6.l)  store the channel number
           move.l   a1,bv_rip(a6)  store the new value of arithmetic stack
           moveq    #3,d4          return a word integer result

no_err     moveq    #0,d0          we have no errors
           rts                     home to basic

*-----------------------------------------------------------------------*
* Insert various library routines (to save typing !)                    *
*-----------------------------------------------------------------------*

           get     "win1_libs_D1TOFP"
           get     "win1_libs_BADPARAM"
           get     "win1_libs_D1TOCHANID"

           END



